"""Tests for _version module."""


def test_version_module() -> None:
    """Tests that _version exports are present."""
    # import kept here to allow mypy/pylint to run when module is not installed
    # and the generated _version.py is missing.
    # pylint: disable=no-name-in-module,no-member
    import ansible_compat._version  # type: ignore[import-not-found,unused-ignore]

    assert ansible_compat._version.__version__
    assert ansible_compat._version.__version_tuple__
    assert ansible_compat._version.version



"""Tests for types module."""

import ansible_compat.types


def test_types() -> None:
    """Tests that JSON types are exported."""
    assert ansible_compat.types.JSON
    assert ansible_compat.types.JSON_ro



"""Tests for schema utilities."""

from __future__ import annotations

import json
from pathlib import Path
from typing import TYPE_CHECKING, Any

import pytest

from ansible_compat.schema import JsonSchemaError, json_path, validate

if TYPE_CHECKING:
    from ansible_compat.types import JSON

expected_results = [
    JsonSchemaError(
        message="False is not of type 'string'",
        data_path="environment.a",
        json_path="$.environment.a",
        schema_path="properties.environment.additionalProperties.type",
        relative_schema='{"type": "string"}',
        expected="string",
        validator="type",
        found="False",
    ),
    JsonSchemaError(
        message="True is not of type 'string'",
        data_path="environment.b",
        json_path="$.environment.b",
        schema_path="properties.environment.additionalProperties.type",
        relative_schema='{"type": "string"}',
        expected="string",
        validator="type",
        found="True",
    ),
]


def json_from_asset(file_name: str) -> JSON:
    """Load a json file from disk."""
    file = Path(__file__).parent / file_name
    with file.open(encoding="utf-8") as f:
        return json.load(f)  # type: ignore[no-any-return]


def jsonify(data: Any) -> JSON:  # noqa: ANN401
    """Convert object in JSON data structure."""
    return json.loads(json.dumps(data, default=vars, sort_keys=True))  # type: ignore[no-any-return]


@pytest.mark.parametrize("index", range(1))
def test_schema(index: int) -> None:
    """Test the schema validator."""
    schema = json_from_asset(f"assets/validate{index}_schema.json")
    data = json_from_asset(f"assets/validate{index}_data.json")
    expected = json_from_asset(f"assets/validate{index}_expected.json")

    # ensure we produce consistent results between runs
    for _ in range(1, 100):
        found_errors = validate(schema=schema, data=data)
        # ensure returned results are already sorted, as we assume our class
        # knows how to sort itself
        assert sorted(found_errors) == found_errors, "multiple errors not sorted"

        found_errors_json = jsonify(found_errors)
        assert (
            found_errors_json == expected
        ), f"inconsistent returns: {found_errors_json}"


def test_json_path() -> None:
    """Test json_path function."""
    assert json_path(["a", 1, "b"]) == "$.a[1].b"


def test_validate_invalid_schema() -> None:
    """Test validate function error handling."""
    schema = "[]"
    data = json_from_asset("assets/validate0_data.json")
    errors = validate(schema, data)

    assert len(errors) == 1
    assert (
        errors[0].to_friendly()
        == "In 'schema sanity check': Invalid schema, must be a mapping."
    )



#!python3
"""Runs downstream projects tests with current code from compat injected in them."""

import hashlib
import logging
import os
import tempfile
from pathlib import Path
from subprocess import run  # noqa: S404

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s: %(message)s",
)
logger = logging.getLogger()


parent_project_dir = Path(__file__).parent.parent.resolve().as_posix()
checksum = hashlib.sha256(parent_project_dir.encode("utf-8")).hexdigest()[:4]
tmp_path = Path(tempfile.gettempdir()) / f"ansible-compat-smoke-{checksum}"

logger.info("Using %s temporary directory...", tmp_path)

for project in ("molecule", "ansible-lint"):

    logger.info("Running tests for %s", project)
    project_dir = tmp_path / project
    if (project_dir / ".git").exists():
        run(["git", "-C", project_dir, "pull"], check=True)
    else:
        project_dir.mkdir(parents=True, exist_ok=True)
        run(
            [
                "git",
                "clone",
                "--recursive",
                f"https://github.com/ansible/{project}",
                project_dir,
            ],
            check=True,
        )

    os.chdir(project_dir)
    venv_dir = (project_dir / ".venv").as_posix()
    os.environ["VIRTUAL_ENV"] = venv_dir
    run(
        ["uv", "venv", "--seed", venv_dir],
        check=True,
    )  # creates .venv (implicit for next commands)
    run(
        ["uv", "pip", "install", "-e", f"{parent_project_dir}[test]", "-e", ".[test]"],
        check=True,
    )
    run(["uv", "pip", "freeze"], check=True)
    run(["uv", "run", "pytest", "-v", "-n", "auto"], check=True)



"""Tests for Runtime class."""

# pylint: disable=protected-access,too-many-lines
from __future__ import annotations

import logging
import os
import pathlib
import subprocess
from contextlib import contextmanager
from pathlib import Path
from shutil import rmtree
from typing import TYPE_CHECKING, Any

import pytest
from packaging.version import Version

from ansible_compat.constants import INVALID_PREREQUISITES_RC
from ansible_compat.errors import (
    AnsibleCommandError,
    AnsibleCompatError,
    InvalidPrerequisiteError,
)
from ansible_compat.runtime import (
    CompletedProcess,
    Runtime,
    _get_galaxy_role_name,
    is_url,
    search_galaxy_paths,
)

if TYPE_CHECKING:
    from collections.abc import Iterator

    from _pytest.monkeypatch import MonkeyPatch
    from pytest_mock import MockerFixture


def test_runtime_version(runtime: Runtime) -> None:
    """Tests version property."""
    version = runtime.version
    assert isinstance(version, Version)
    # tests that caching property value worked (coverage)
    assert version == runtime.version


@pytest.mark.parametrize(
    "require_module",
    (True, False),
    ids=("module-required", "module-unrequired"),
)
def test_runtime_version_outdated(require_module: bool) -> None:
    """Checks that instantiation raises if version is outdated."""
    with pytest.raises(RuntimeError, match="Found incompatible version of ansible"):
        Runtime(min_required_version="9999.9.9", require_module=require_module)


def test_runtime_missing_ansible_module(monkeypatch: MonkeyPatch) -> None:
    """Checks that we produce a RuntimeError when ansible module is missing."""

    class RaiseException:
        """Class to raise an exception."""

        def __init__(
            self,
            *args: Any,  # noqa: ARG002,ANN401
            **kwargs: Any,  # noqa: ARG002,ANN401
        ) -> None:
            raise ModuleNotFoundError

    monkeypatch.setattr("importlib.import_module", RaiseException)

    with pytest.raises(RuntimeError, match=r"Unable to find Ansible python module."):
        Runtime(require_module=True)


def test_runtime_mismatch_ansible_module(monkeypatch: MonkeyPatch) -> None:
    """Test that missing module is detected."""
    monkeypatch.setattr("ansible.release.__version__", "0.0.0", raising=False)
    with pytest.raises(RuntimeError, match="versions do not match"):
        Runtime(require_module=True)


def test_runtime_require_module() -> None:
    """Check that require_module successful pass."""
    Runtime(require_module=True)
    # Now we try to set the collection path, something to check if that is
    # causing an exception, as 2.15 introduced new init code.
    from ansible.utils.collection_loader import (  # pylint: disable=import-outside-toplevel
        AnsibleCollectionConfig,
    )

    AnsibleCollectionConfig.playbook_paths = "."
    # Calling it again in order to see that it does not produce UserWarning: AnsibleCollectionFinder has already been configured
    # which is done by Ansible core 2.15+. We added special code inside Runtime
    # that should avoid initializing twice and raise that warning.
    Runtime(require_module=True)


def test_runtime_version_fail_module(mocker: MockerFixture) -> None:
    """Tests for failure to detect Ansible version."""
    patched = mocker.patch(
        "ansible_compat.runtime.parse_ansible_version",
        autospec=True,
    )
    patched.side_effect = InvalidPrerequisiteError(
        "Unable to parse ansible cli version",
    )
    runtime = Runtime()
    with pytest.raises(
        InvalidPrerequisiteError,
        match="Unable to parse ansible cli version",
    ):
        _ = runtime.version  # pylint: disable=pointless-statement


def test_runtime_version_fail_cli(mocker: MockerFixture) -> None:
    """Tests for failure to detect Ansible version."""
    mocker.patch(
        "ansible_compat.runtime.Runtime.run",
        return_value=CompletedProcess(
            ["x"],
            returncode=123,
            stdout="oops",
            stderr="some error",
        ),
        autospec=True,
    )
    runtime = Runtime()
    with pytest.raises(
        RuntimeError,
        match=r"Unable to find a working copy of ansible executable.",
    ):
        _ = runtime.version  # pylint: disable=pointless-statement


def test_runtime_prepare_ansible_paths_validation() -> None:
    """Check that we validate collection_path."""
    runtime = Runtime()
    runtime.config.collections_paths = "invalid-value"  # type: ignore[assignment]
    with pytest.raises(RuntimeError, match="Unexpected ansible configuration"):
        runtime._prepare_ansible_paths()


@pytest.mark.parametrize(
    ("folder", "role_name", "isolated"),
    (
        ("ansible-role-sample", "acme.sample", True),
        ("acme.sample2", "acme.sample2", True),
        ("sample3", "acme.sample3", True),
        ("sample4", "acme.sample4", False),
    ),
    ids=("1", "2", "3", "4"),
)
def test_runtime_install_role(
    caplog: pytest.LogCaptureFixture,
    folder: str,
    role_name: str,
    isolated: bool,
) -> None:
    """Checks that we can install roles."""
    caplog.set_level(logging.INFO)
    project_dir = Path(__file__).parent / "roles" / folder
    runtime = Runtime(isolated=isolated, project_dir=project_dir)
    runtime.prepare_environment(install_local=True)
    # check that role appears as installed now
    result = runtime.run(["ansible-galaxy", "list"])
    assert result.returncode == 0, result
    assert role_name in result.stdout
    if isolated:
        assert pathlib.Path(f"{runtime.cache_dir}/roles/{role_name}").is_symlink()
    else:
        assert pathlib.Path(
            f"{Path(runtime.config.default_roles_path[0]).expanduser()}/{role_name}",
        ).is_symlink()
    runtime.clean()


def test_prepare_environment_with_collections(runtime_tmp: Runtime) -> None:
    """Check that collections are correctly installed."""
    runtime_tmp.prepare_environment(
        required_collections={"community.molecule": "0.1.0"},
        install_local=True,
    )
    assert "community.molecule" in runtime_tmp.collections


def test_runtime_install_requirements_missing_file() -> None:
    """Check that missing requirements file is ignored."""
    # Do not rely on this behavior, it may be removed in the future
    runtime = Runtime()
    runtime.install_requirements(Path("/that/does/not/exist"))


@pytest.mark.parametrize(
    ("file", "exc", "msg"),
    (
        (
            Path("/dev/null"),
            InvalidPrerequisiteError,
            "file is not a valid Ansible requirements file",
        ),
        (
            Path(__file__).parent / "assets" / "requirements-invalid-collection.yml",
            AnsibleCommandError,
            "Got 1 exit code while running: ansible-galaxy",
        ),
        (
            Path(__file__).parent / "assets" / "requirements-invalid-role.yml",
            AnsibleCommandError,
            "Got 1 exit code while running: ansible-galaxy",
        ),
    ),
    ids=("empty", "invalid-collection", "invalid-role"),
)
def test_runtime_install_requirements_invalid_file(
    file: Path,
    exc: type[Any],
    msg: str,
) -> None:
    """Check that invalid requirements file is raising."""
    runtime = Runtime()
    with pytest.raises(
        exc,
        match=msg,
    ):
        runtime.install_requirements(file)


@contextmanager
def cwd(path: Path) -> Iterator[None]:
    """Context manager for temporary changing current working directory."""
    old_pwd = Path.cwd()
    os.chdir(path)
    try:
        yield
    finally:
        os.chdir(old_pwd)


def test_prerun_reqs_v1(caplog: pytest.LogCaptureFixture) -> None:
    """Checks that the linter can auto-install requirements v1 when found."""
    path = Path(__file__).parent.parent / "examples" / "reqs_v1"
    runtime = Runtime(project_dir=path, verbosity=1)
    with cwd(path):
        runtime.prepare_environment()
    assert any(
        msg.startswith("Running ansible-galaxy role install") for msg in caplog.messages
    )
    assert all(
        "Running ansible-galaxy collection install" not in msg
        for msg in caplog.messages
    )


def test_prerun_reqs_v2(caplog: pytest.LogCaptureFixture) -> None:
    """Checks that the linter can auto-install requirements v2 when found."""
    path = (Path(__file__).parent.parent / "examples" / "reqs_v2").resolve()
    runtime = Runtime(project_dir=path, verbosity=1)
    with cwd(path):
        runtime.prepare_environment()
        assert any(
            msg.startswith("Running ansible-galaxy role install")
            for msg in caplog.messages
        )
        assert any(
            msg.startswith("Running ansible-galaxy collection install")
            for msg in caplog.messages
        )


def test_prerun_reqs_broken() -> None:
    """Checks that the we report invalid requirements.yml file."""
    path = (Path(__file__).parent.parent / "examples" / "reqs_broken").resolve()
    runtime = Runtime(project_dir=path, verbosity=1)
    with cwd(path), pytest.raises(InvalidPrerequisiteError):
        runtime.prepare_environment()


def test__update_env_no_old_value_no_default_no_value(monkeypatch: MonkeyPatch) -> None:
    """Make sure empty value does not touch environment."""
    monkeypatch.delenv("DUMMY_VAR", raising=False)

    runtime = Runtime()
    runtime._update_env("DUMMY_VAR", [])

    assert "DUMMY_VAR" not in runtime.environ


def test__update_env_no_old_value_no_value(monkeypatch: MonkeyPatch) -> None:
    """Make sure empty value does not touch environment."""
    monkeypatch.delenv("DUMMY_VAR", raising=False)

    runtime = Runtime()
    runtime._update_env("DUMMY_VAR", [], "a:b")

    assert "DUMMY_VAR" not in runtime.environ


def test__update_env_no_default_no_value(monkeypatch: MonkeyPatch) -> None:
    """Make sure empty value does not touch environment."""
    monkeypatch.setenv("DUMMY_VAR", "a:b")

    runtime = Runtime()
    runtime._update_env("DUMMY_VAR", [])

    assert runtime.environ["DUMMY_VAR"] == "a:b"


@pytest.mark.parametrize(
    ("value", "result"),
    (
        (["a"], "a"),
        (["a", "b"], "a:b"),
        (["a", "b", "c"], "a:b:c"),
    ),
)
def test__update_env_no_old_value_no_default(
    monkeypatch: MonkeyPatch,
    value: list[str],
    result: str,
) -> None:
    """Values are concatenated using : as the separator."""
    monkeypatch.delenv("DUMMY_VAR", raising=False)

    runtime = Runtime()
    runtime._update_env("DUMMY_VAR", value)

    assert runtime.environ["DUMMY_VAR"] == result


@pytest.mark.parametrize(
    ("default", "value", "result"),
    (
        ("a:b", ["c"], "c:a:b"),
        ("a:b", ["c:d"], "c:d:a:b"),
    ),
)
def test__update_env_no_old_value(
    monkeypatch: MonkeyPatch,
    default: str,
    value: list[str],
    result: str,
) -> None:
    """Values are appended to default value."""
    monkeypatch.delenv("DUMMY_VAR", raising=False)

    runtime = Runtime()
    runtime._update_env("DUMMY_VAR", value, default)

    assert runtime.environ["DUMMY_VAR"] == result


@pytest.mark.parametrize(
    ("old_value", "value", "result"),
    (
        ("a:b", ["c"], "c:a:b"),
        ("a:b", ["c:d"], "c:d:a:b"),
    ),
)
def test__update_env_no_default(
    monkeypatch: MonkeyPatch,
    old_value: str,
    value: list[str],
    result: str,
) -> None:
    """Values are appended to preexisting value."""
    monkeypatch.setenv("DUMMY_VAR", old_value)

    runtime = Runtime()
    runtime._update_env("DUMMY_VAR", value)

    assert runtime.environ["DUMMY_VAR"] == result


@pytest.mark.parametrize(
    ("old_value", "default", "value", "result"),
    (
        ("", "", ["e"], "e"),
        ("a", "", ["e"], "e:a"),
        ("", "c", ["e"], "e"),
        ("a", "c", ["e:f"], "e:f:a"),
    ),
)
def test__update_env(
    monkeypatch: MonkeyPatch,
    old_value: str,
    default: str,  # pylint: disable=unused-argument # noqa: ARG001
    value: list[str],
    result: str,
) -> None:
    """Defaults are ignored when preexisting value is present."""
    monkeypatch.setenv("DUMMY_VAR", old_value)

    runtime = Runtime()
    runtime._update_env("DUMMY_VAR", value)

    assert runtime.environ["DUMMY_VAR"] == result


def test_require_collection_wrong_version(runtime: Runtime) -> None:
    """Tests behavior of require_collection."""
    subprocess.check_output(
        [
            "ansible-galaxy",
            "collection",
            "install",
            "examples/reqs_v2/community-molecule-0.1.0.tar.gz",
            "-p",
            "~/.ansible/collections",
        ],
    )
    with pytest.raises(InvalidPrerequisiteError) as pytest_wrapped_e:
        runtime.require_collection("community.molecule", "9999.9.9")
    assert pytest_wrapped_e.type == InvalidPrerequisiteError
    assert pytest_wrapped_e.value.code == INVALID_PREREQUISITES_RC


def test_require_collection_invalid_name(runtime: Runtime) -> None:
    """Check that require_collection raise with invalid collection name."""
    with pytest.raises(
        InvalidPrerequisiteError,
        match="Invalid collection name supplied:",
    ):
        runtime.require_collection("that-is-invalid")


def test_require_collection_invalid_collections_path(runtime: Runtime) -> None:
    """Check that require_collection raise with invalid collections path."""
    runtime.config.collections_paths = "/that/is/invalid"  # type: ignore[assignment]
    with pytest.raises(
        InvalidPrerequisiteError,
        match="Unable to determine ansible collection paths",
    ):
        runtime.require_collection("community.molecule")


def test_require_collection_preexisting_broken(runtime_tmp: Runtime) -> None:
    """Check that require_collection raise with broken pre-existing collection."""
    dest_path: str = runtime_tmp.config.collections_paths[0]
    dest = pathlib.Path(dest_path) / "ansible_collections" / "foo" / "bar"
    dest.mkdir(parents=True, exist_ok=True)
    with pytest.raises(InvalidPrerequisiteError, match=r"missing MANIFEST.json"):
        runtime_tmp.require_collection("foo.bar")


def test_require_collection_install(runtime_tmp: Runtime) -> None:
    """Check that require collection successful install case, including upgrade path."""
    runtime_tmp.install_collection("ansible.posix:==1.5.2")
    runtime_tmp.load_collections()
    collection = runtime_tmp.collections["ansible.posix"]
    assert collection.version == "1.5.2"
    runtime_tmp.require_collection(name="ansible.posix", version="1.5.4", install=True)
    runtime_tmp.load_collections()
    collection = runtime_tmp.collections["ansible.posix"]
    assert Version(collection.version) >= Version("1.5.4")


@pytest.mark.parametrize(
    ("name", "version", "install"),
    (
        ("fake_namespace.fake_name", None, True),
        ("fake_namespace.fake_name", "9999.9.9", True),
        ("fake_namespace.fake_name", None, False),
    ),
    ids=("a", "b", "c"),
)
def test_require_collection_missing(
    name: str,
    version: str,
    install: bool,
    runtime: Runtime,
) -> None:
    """Tests behavior of require_collection, missing case."""
    with pytest.raises(AnsibleCompatError) as pytest_wrapped_e:
        runtime.require_collection(name=name, version=version, install=install)
    assert pytest_wrapped_e.type == InvalidPrerequisiteError
    assert pytest_wrapped_e.value.code == INVALID_PREREQUISITES_RC


def test_install_collection(runtime: Runtime) -> None:
    """Check that valid collection installs do not fail."""
    runtime.install_collection("examples/reqs_v2/community-molecule-0.1.0.tar.gz")


def test_install_collection_git(runtime: Runtime) -> None:
    """Check that valid collection installs do not fail."""
    runtime.install_collection(
        "git+https://github.com/ansible-collections/ansible.posix,main",
    )


def test_install_collection_dest(runtime: Runtime, tmp_path: pathlib.Path) -> None:
    """Check that valid collection to custom destination passes."""
    # Since Ansible 2.15.3 there is no guarantee that this will install the collection at requested path
    # as it might decide to not install anything if requirement is already present at another location.
    runtime.install_collection(
        "examples/reqs_v2/community-molecule-0.1.0.tar.gz",
        destination=tmp_path,
    )
    runtime.load_collections()
    for collection in runtime.collections:
        if collection == "community.molecule":
            return
    msg = "Failed to find collection as installed."
    raise AssertionError(msg)


def test_install_collection_fail(runtime: Runtime) -> None:
    """Check that invalid collection install fails."""
    with pytest.raises(AnsibleCompatError) as pytest_wrapped_e:
        runtime.install_collection("community.molecule:>=9999.0")
    assert pytest_wrapped_e.type == InvalidPrerequisiteError
    assert pytest_wrapped_e.value.code == INVALID_PREREQUISITES_RC


def test_install_galaxy_role(runtime_tmp: Runtime) -> None:
    """Check install role with empty galaxy file."""
    pathlib.Path(f"{runtime_tmp.project_dir}/galaxy.yml").touch()
    pathlib.Path(f"{runtime_tmp.project_dir}/meta").mkdir()
    pathlib.Path(f"{runtime_tmp.project_dir}/meta/main.yml").touch()
    # this should only raise a warning
    runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=1)
    # this should test the bypass role name check path
    runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=2)
    # this should raise an error
    with pytest.raises(
        InvalidPrerequisiteError,
        match="does not follow current galaxy requirements",
    ):
        runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=0)


def test_install_galaxy_role_unlink(
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Test ability to unlink incorrect symlinked roles."""
    runtime_tmp = Runtime(verbosity=1, isolated=True)
    runtime_tmp.prepare_environment()
    assert runtime_tmp.cache_dir is not None
    pathlib.Path(f"{runtime_tmp.cache_dir}/roles").mkdir(parents=True, exist_ok=True)
    roledir = pathlib.Path(f"{runtime_tmp.cache_dir}/roles/acme.get_rich")
    if not roledir.exists():
        roledir.symlink_to("/dev/null")
    pathlib.Path(f"{runtime_tmp.project_dir}/meta").mkdir(exist_ok=True)
    pathlib.Path(f"{runtime_tmp.project_dir}/meta/main.yml").write_text(
        """galaxy_info:
  role_name: get_rich
  namespace: acme
""",
        encoding="utf-8",
    )
    runtime_tmp._install_galaxy_role(runtime_tmp.project_dir)
    assert "symlink to current repository" in caplog.text
    pathlib.Path(f"{runtime_tmp.project_dir}/meta/main.yml").unlink()


def test_install_galaxy_role_bad_namespace(runtime_tmp: Runtime) -> None:
    """Check install role with bad namespace in galaxy info."""
    pathlib.Path(f"{runtime_tmp.project_dir}/meta").mkdir()
    pathlib.Path(f"{runtime_tmp.project_dir}/meta/main.yml").write_text(
        """galaxy_info:
  role_name: foo
  author: bar
  namespace: ["xxx"]
""",
        encoding="utf-8",
    )
    # this should raise an error regardless the role_name_check value
    with pytest.raises(AnsibleCompatError, match="Role namespace must be string, not"):
        runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=1)


def test_install_galaxy_role_no_meta(runtime_tmp: Runtime) -> None:
    """Check install role with missing meta/main.yml."""
    # This should fail because meta/main.yml is missing
    with pytest.raises(
        FileNotFoundError,
        match=f"No such file or directory: '{runtime_tmp.project_dir.absolute()}/meta/main.yaml'",
    ):
        runtime_tmp._install_galaxy_role(runtime_tmp.project_dir)
    # But ignore_errors will return without doing anything
    runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, ignore_errors=True)


@pytest.mark.parametrize(
    "galaxy_info",
    (
        """galaxy_info:
  role_name: foo-bar
  namespace: acme
""",
        """galaxy_info:
  role_name: foo-bar
""",
    ),
    ids=("bad-name", "bad-name-without-namespace"),
)
def test_install_galaxy_role_name_role_name_check_equals_to_1(
    runtime_tmp: Runtime,
    galaxy_info: str,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Check install role with bad role name in galaxy info."""
    caplog.set_level(logging.WARNING)
    pathlib.Path(f"{runtime_tmp.project_dir}/meta").mkdir()
    pathlib.Path(f"{runtime_tmp.project_dir}/meta/main.yml").write_text(
        galaxy_info,
        encoding="utf-8",
    )

    runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=1)
    assert "Computed fully qualified role name of " in caplog.text


def test_install_galaxy_role_no_checks(runtime_tmp: Runtime) -> None:
    """Check install role with bad namespace in galaxy info."""
    runtime_tmp.prepare_environment()
    pathlib.Path(f"{runtime_tmp.project_dir}/meta").mkdir()
    pathlib.Path(f"{runtime_tmp.project_dir}/meta/main.yml").write_text(
        """galaxy_info:
  role_name: foo
  author: bar
  namespace: acme
""",
        encoding="utf-8",
    )
    runtime_tmp._install_galaxy_role(runtime_tmp.project_dir, role_name_check=2)
    result = runtime_tmp.run(["ansible-galaxy", "list"])
    assert "- acme.foo," in result.stdout
    assert result.returncode == 0, result


def test_upgrade_collection(runtime_tmp: Runtime) -> None:
    """Check that collection upgrade is possible."""
    # ensure that we inject our tmp folders in ansible paths
    runtime_tmp.prepare_environment()

    # we install specific outdated version of a collection
    runtime_tmp.install_collection("examples/reqs_v2/community-molecule-0.1.0.tar.gz")
    with pytest.raises(
        InvalidPrerequisiteError,
        match=r"Found community.molecule collection 0.1.0 but 9.9.9 or newer is required.",
    ):
        # we check that when install=False, we raise error
        runtime_tmp.require_collection("community.molecule", "9.9.9", install=False)
    # this should not fail, as we have this version
    runtime_tmp.require_collection("community.molecule", "0.1.0")


def test_require_collection_not_isolated() -> None:
    """Check require_collection without a cache directory."""
    runtime = Runtime(isolated=False)
    runtime.require_collection("community.molecule", "0.1.0", install=True)


def test_runtime_env_ansible_library(monkeypatch: MonkeyPatch) -> None:
    """Verify that custom path specified using ANSIBLE_LIBRARY is not lost."""
    path_name = "foo"
    monkeypatch.setenv("ANSIBLE_LIBRARY", path_name)

    path_name = os.path.realpath(path_name)
    runtime = Runtime()
    runtime.prepare_environment()
    assert path_name in runtime.config.default_module_path


@pytest.mark.parametrize(
    ("lower", "upper", "expected"),
    (
        ("1.0", "9999.0", True),
        (None, "9999.0", True),
        ("1.0", None, True),
        ("9999.0", None, False),
        (None, "1.0", False),
    ),
    ids=("1", "2", "3", "4", "5"),
)
def test_runtime_version_in_range(
    lower: str | None,
    upper: str | None,
    expected: bool,
) -> None:
    """Validate functioning of version_in_range."""
    runtime = Runtime()
    assert runtime.version_in_range(lower=lower, upper=upper) is expected


@pytest.mark.parametrize(
    ("path", "scenario", "expected_collections"),
    (
        pytest.param(
            "test/collections/acme.goodies",
            "default",
            [
                "ansible.posix",  # from tests/requirements.yml
                "ansible.utils",  # from galaxy.yml
                "community.molecule",  # from galaxy.yml
                "community.crypto",  # from galaxy.yml as a git dependency
            ],
            id="normal",
        ),
        pytest.param(
            "test/collections/acme.goodies/roles/baz",
            "deep_scenario",
            ["community.molecule"],
            id="deep",
        ),
    ),
)
def test_install_collection_from_disk(
    path: str,
    scenario: str,
    expected_collections: list[str],
) -> None:
    """Tests ability to install a local collection."""
    # ensure we do not have acme.goodies installed in user directory as it may
    # produce false positives
    rmtree(
        pathlib.Path(
            "~/.ansible/collections/ansible_collections/acme/goodies",
        ).expanduser(),
        ignore_errors=True,
    )
    with cwd(Path(path)):
        runtime = Runtime(isolated=True)
        # this should call install_collection_from_disk(".")
        runtime.prepare_environment(install_local=True)
        # that molecule converge playbook can be used without molecule and
        # should validate that the installed collection is available.
        result = runtime.run(["ansible-playbook", f"molecule/{scenario}/converge.yml"])
        assert result.returncode == 0, result.stdout
        runtime.load_collections()
        for collection_name in expected_collections:
            assert (
                collection_name in runtime.collections
            ), f"{collection_name} not found in {runtime.collections.keys()}"
        runtime.clean()


@pytest.mark.parametrize(
    ("path", "expected_plugins"),
    (
        pytest.param(
            "test/collections/acme.goodies",
            [
                "ansible.posix.patch",  # from tests/requirements.yml
                "community.crypto.acme_account",  # from galaxy.yml as a git dependency
            ],
            id="modules",
        ),
    ),
)
def test_load_plugins(
    path: str,
    expected_plugins: list[str],
) -> None:
    """Tests ability to load plugin from a collection installed by requirement."""
    with cwd(Path(path)):
        runtime = Runtime(isolated=True, require_module=True)
        runtime.prepare_environment(install_local=True)
        for plugin_name in expected_plugins:
            assert (
                plugin_name in runtime.plugins.module
            ), f"Unable to load module {plugin_name}"

        runtime.clean()


def test_install_collection_from_disk_fail() -> None:
    """Tests that we fail to install a broken collection."""
    with cwd(Path("test/collections/acme.broken")):
        runtime = Runtime(isolated=True)
        with pytest.raises(RuntimeError) as exc_info:
            runtime.prepare_environment(install_local=True)
        # based on version of Ansible used, we might get a different error,
        # but both errors should be considered acceptable
        assert exc_info.type in {
            RuntimeError,
            AnsibleCompatError,
            AnsibleCommandError,
            InvalidPrerequisiteError,
        }
        assert exc_info.match(
            "(is missing the following mandatory|Got 1 exit code while running: ansible-galaxy collection build)",
        )


def test_load_collections_failure(mocker: MockerFixture) -> None:
    """Tests for ansible-galaxy erroring."""
    mocker.patch(
        "ansible_compat.runtime.Runtime.run",
        return_value=CompletedProcess(
            ["x"],
            returncode=1,
            stdout="There was an error",
            stderr="This is the error",
        ),
        autospec=True,
    )
    runtime = Runtime()
    with pytest.raises(RuntimeError, match="Unable to list collections: "):
        runtime.load_collections()


@pytest.mark.parametrize(
    "value",
    ("[]", '{"path": "bad data"}', '{"path": {"ansible.posix": 123}}'),
    ids=["list", "malformed_collection", "bad_collection_data"],
)
def test_load_collections_garbage(value: str, mocker: MockerFixture) -> None:
    """Tests for ansible-galaxy returning bad data."""
    mocker.patch(
        "ansible_compat.runtime.Runtime.run",
        return_value=CompletedProcess(
            ["x"],
            returncode=0,
            stdout=value,
            stderr="",
        ),
        autospec=True,
    )
    runtime = Runtime()
    with pytest.raises(TypeError, match="Unexpected collection data, "):
        runtime.load_collections()


@pytest.mark.parametrize(
    "value",
    ("", '{"path": {123: 456}}'),
    ids=["nothing", "bad_collection_name"],
)
def test_load_collections_invalid_json(value: str, mocker: MockerFixture) -> None:
    """Tests for ansible-galaxy returning bad data."""
    mocker.patch(
        "ansible_compat.runtime.Runtime.run",
        return_value=CompletedProcess(
            ["x"],
            returncode=0,
            stdout=value,
            stderr="",
        ),
        autospec=True,
    )
    runtime = Runtime()
    with pytest.raises(
        RuntimeError,
        match=f"Unable to parse galaxy output as JSON: {value}",
    ):
        runtime.load_collections()


def test_prepare_environment_offline_role(caplog: pytest.LogCaptureFixture) -> None:
    """Ensure that we can make use of offline roles."""
    with cwd(Path("test/roles/acme.missing_deps")):
        runtime = Runtime(isolated=True)
        runtime.prepare_environment(install_local=True, offline=True)
        assert (
            "Skipped installing old role dependencies due to running in offline mode."
            in caplog.text
        )
        assert (
            "Skipped installing collection dependencies due to running in offline mode."
            in caplog.text
        )


def test_runtime_run(runtime: Runtime) -> None:
    """Check if tee and non tee mode return same kind of results."""
    result1 = runtime.run(["seq", "10"])
    result2 = runtime.run(["seq", "10"], tee=True)
    assert result1.returncode == result2.returncode
    assert result1.stderr == result2.stderr
    assert result1.stdout == result2.stdout


def test_runtime_exec_cwd(runtime: Runtime) -> None:
    """Check if passing cwd works as expected."""
    path = Path("/")
    result1 = runtime.run(["pwd"], cwd=path)
    result2 = runtime.run(["pwd"])
    assert result1.stdout.rstrip() == str(path)
    assert result1.stdout != result2.stdout


def test_runtime_exec_env(runtime: Runtime) -> None:
    """Check if passing env works."""
    result = runtime.run(["printenv", "FOO"])
    assert not result.stdout

    result = runtime.run(["printenv", "FOO"], env={"FOO": "bar"})
    assert result.stdout.rstrip() == "bar"

    runtime.environ["FOO"] = "bar"
    result = runtime.run(["printenv", "FOO"])
    assert result.stdout.rstrip() == "bar"


def test_runtime_plugins(runtime: Runtime) -> None:
    """Tests ability to access detected plugins."""
    assert len(runtime.plugins.cliconf) == 0
    # ansible.netcommon.restconf might be in httpapi
    assert isinstance(runtime.plugins.httpapi, dict)
    # "ansible.netcommon.default" might be in runtime.plugins.netconf
    assert isinstance(runtime.plugins.netconf, dict)
    assert isinstance(runtime.plugins.role, dict)
    assert "become" in runtime.plugins.keyword

    assert "ansible.builtin.sudo" in runtime.plugins.become
    assert "ansible.builtin.memory" in runtime.plugins.cache
    assert "ansible.builtin.default" in runtime.plugins.callback
    assert "ansible.builtin.local" in runtime.plugins.connection
    assert "ansible.builtin.ini" in runtime.plugins.inventory
    assert "ansible.builtin.env" in runtime.plugins.lookup
    assert "ansible.builtin.sh" in runtime.plugins.shell
    assert "ansible.builtin.host_group_vars" in runtime.plugins.vars
    assert "ansible.builtin.file" in runtime.plugins.module
    assert "ansible.builtin.free" in runtime.plugins.strategy
    assert "ansible.builtin.is_abs" in runtime.plugins.test
    assert "ansible.builtin.bool" in runtime.plugins.filter


@pytest.mark.parametrize(
    ("path", "result"),
    (
        pytest.param(
            Path("test/assets/galaxy_paths"),
            [Path("test/assets/galaxy_paths/foo/galaxy.yml").resolve()],
            id="1",
        ),
        pytest.param(
            Path("test/collections"),
            [],  # should find nothing because these folders are not valid namespaces
            id="2",
        ),
        pytest.param(
            Path("test/assets/galaxy_paths/foo"),
            [Path("test/assets/galaxy_paths/foo/galaxy.yml").resolve()],
            id="3",
        ),
    ),
)
def test_galaxy_path(path: Path, result: list[Path]) -> None:
    """Check behavior of galaxy path search."""
    assert search_galaxy_paths(path) == result


@pytest.mark.parametrize(
    ("name", "result"),
    (
        pytest.param(
            "foo",
            False,
            id="0",
        ),
        pytest.param(
            "git+git",
            True,
            id="1",
        ),
        pytest.param(
            "git@acme.com",
            True,
            id="2",
        ),
    ),
)
def test_is_url(name: str, result: bool) -> None:
    """Checks functionality of is_url."""
    assert is_url(name) == result


@pytest.mark.parametrize(
    ("dest", "message"),
    (
        ("/invalid/destination", "Collection is symlinked, but not pointing to"),
        (Path.cwd(), "Found symlinked collection, skipping its installation."),
    ),
    ids=["broken", "valid"],
)
def test_prepare_environment_symlink(
    dest: str | Path,
    message: str,
    caplog: pytest.LogCaptureFixture,
) -> None:
    """Ensure avalid symlinks to collections are properly detected."""
    project_dir = Path(__file__).parent / "collections" / "acme.minimal"
    runtime = Runtime(isolated=True, project_dir=project_dir)
    assert runtime.cache_dir
    acme = runtime.cache_dir / "collections" / "ansible_collections" / "acme"
    acme.mkdir(parents=True, exist_ok=True)
    goodies = acme / "minimal"
    rmtree(goodies, ignore_errors=True)
    goodies.unlink(missing_ok=True)
    goodies.symlink_to(dest)
    runtime.prepare_environment(install_local=True)
    assert message in caplog.text


def test_get_galaxy_role_name_invalid() -> None:
    """Verifies that function returns empty string on invalid input."""
    galaxy_infos = {
        "role_name": False,  # <-- invalid data, should be string
    }
    assert not _get_galaxy_role_name(galaxy_infos)


def test_runtime_has_playbook() -> None:
    """Tests has_playbook method."""
    runtime = Runtime(require_module=True)

    runtime.prepare_environment(
        required_collections={"community.molecule": "0.1.0"},
        install_local=True,
    )

    assert not runtime.has_playbook("this-does-not-exist.yml")
    # call twice to ensure cache is used:
    assert not runtime.has_playbook("this-does-not-exist.yml")

    assert not runtime.has_playbook("this-does-not-exist.yml", basedir=Path())
    # this is part of community.molecule collection
    assert runtime.has_playbook("community.molecule.validate.yml")


def test_runtime_exception(monkeypatch: pytest.MonkeyPatch) -> None:
    """Asserts that we raise a runtime exception if unsupported environment variable is detected."""
    monkeypatch.setenv("ANSIBLE_COLLECTIONS_PATHS", "foo")
    with pytest.raises(
        RuntimeError,
        match=r"ANSIBLE_COLLECTIONS_PATHS was detected, replace it with ANSIBLE_COLLECTIONS_PATH to continue.",
    ):
        Runtime()



"""Test the scan path functionality of the runtime."""

import json
import os
import subprocess
import textwrap
from pathlib import Path

import pytest
from _pytest.monkeypatch import MonkeyPatch

from ansible_compat.runtime import Runtime

from .conftest import VirtualEnvironment

V2_COLLECTION_TARBALL = Path("examples/reqs_v2/community-molecule-0.1.0.tar.gz")
V2_COLLECTION_NAMESPACE = "community"
V2_COLLECTION_NAME = "molecule"
V2_COLLECTION_VERSION = "0.1.0"
V2_COLLECTION_FULL_NAME = f"{V2_COLLECTION_NAMESPACE}.{V2_COLLECTION_NAME}"


@pytest.mark.parametrize(
    ("scan", "raises_not_found"),
    (
        pytest.param(False, True, id="disabled"),
        pytest.param(True, False, id="enabled"),
    ),
    ids=str,
)
def test_scan_sys_path(
    venv_module: VirtualEnvironment,
    monkeypatch: MonkeyPatch,
    tmp_path: Path,
    scan: bool,
    raises_not_found: bool,
) -> None:
    """Confirm sys path is scanned for collections.

    Args:
        venv_module: Fixture for a virtual environment
        monkeypatch: Fixture for monkeypatching
        tmp_path: Fixture for a temporary directory
        scan: Whether to scan the sys path
        raises_not_found: Whether the collection is expected to be found
    """
    # Isolated the test from the others, so ansible will not find collections
    # that might be installed by other tests.
    monkeypatch.setenv("VIRTUAL_ENV", venv_module.project.as_posix())
    monkeypatch.setenv("ANSIBLE_HOME", tmp_path.as_posix())
    # Set the sys scan path environment variable
    monkeypatch.setenv("ANSIBLE_COLLECTIONS_SCAN_SYS_PATH", str(scan))
    # Set the ansible collections paths to avoid bleed from other tests
    monkeypatch.setenv("ANSIBLE_COLLECTIONS_PATH", str(tmp_path))

    runtime_tmp = Runtime(project_dir=tmp_path, isolated=True)
    first_site_package_dir = venv_module.site_package_dirs()[0]

    installed_to = (
        first_site_package_dir
        / "ansible_collections"
        / V2_COLLECTION_NAMESPACE
        / V2_COLLECTION_NAME
    )
    if not installed_to.exists():
        # Install the collection into the venv site packages directory, force
        # as of yet this test is not isolated from the rest of the system
        runtime_tmp.install_collection(
            collection=V2_COLLECTION_TARBALL,
            destination=first_site_package_dir,
            force=True,
        )
    # Confirm the collection is installed
    assert installed_to.exists()

    script = textwrap.dedent(
        f"""
    import json;
    from ansible_compat.runtime import Runtime;
    r = Runtime();
    fv, cp = r.require_collection(name="{V2_COLLECTION_FULL_NAME}", version="{V2_COLLECTION_VERSION}", install=False);
    print(json.dumps({{"found_version": str(fv), "collection_path": str(cp)}}));
    """,
    )

    proc = venv_module.python_script_run(script)
    if raises_not_found:
        assert proc.returncode != 0, (proc.stdout, proc.stderr)
        assert "InvalidPrerequisiteError" in proc.stderr
        assert "'community.molecule' not found" in proc.stderr
    else:
        assert proc.returncode == 0, (proc.stdout, proc.stderr)
        result = json.loads(proc.stdout)
        assert result["found_version"] == V2_COLLECTION_VERSION
        assert result["collection_path"] == str(installed_to)

    runtime_tmp.clean()


def test_ro_venv() -> None:
    """Tests behavior when the virtual environment is read-only.

    See Related https://github.com/ansible/ansible-compat/pull/470
    """
    tox_work_dir = os.environ.get("TOX_WORK_DIR", ".tox")
    venv_path = f"{tox_work_dir}/ro"
    commands = [
        f"mkdir -p {venv_path}",
        f"chmod -R a+w {venv_path}",
        f"python -m venv --symlinks {venv_path}",
        f"{venv_path}/bin/python -m pip install -q -e .",
        f"chmod -R a-w {venv_path}",
        f"{venv_path}/bin/python -c \"from ansible_compat.runtime import Runtime; r = Runtime(); r.install_collection('ansible.posix:>=2.0.0')\"",
    ]
    for cmd in commands:
        result = subprocess.run(  # noqa: S602
            cmd,
            check=False,
            shell=True,
            text=True,
            capture_output=True,
        )
        assert (
            result.returncode == 0
        ), f"Got {result.returncode} running {cmd}\n\tstderr: {result.stderr}\n\tstdout: {result.stdout}"



"""Tests for ansible_compat.prerun module."""

from __future__ import annotations

import tempfile
from pathlib import Path
from typing import TYPE_CHECKING

import pytest

if TYPE_CHECKING:
    from _pytest.monkeypatch import MonkeyPatch

from ansible_compat.prerun import get_cache_dir


def test_get_cache_dir_relative() -> None:
    """Test behaviors of get_cache_dir."""
    relative_path = Path()
    abs_path = relative_path.resolve()
    assert get_cache_dir(relative_path) == get_cache_dir(abs_path)


def test_get_cache_dir_no_isolation_no_venv(monkeypatch: MonkeyPatch) -> None:
    """Test behaviors of get_cache_dir.

    Args:
        monkeypatch: Pytest fixture for monkeypatching
    """
    monkeypatch.delenv("VIRTUAL_ENV", raising=False)
    monkeypatch.delenv("ANSIBLE_HOME", raising=False)
    assert get_cache_dir(Path(), isolated=False) == Path("~/.ansible").expanduser()


def test_get_cache_dir_isolation_no_venv(monkeypatch: MonkeyPatch) -> None:
    """Test behaviors of get_cache_dir.

    Args:
        monkeypatch: Pytest fixture for monkeypatching
    """
    monkeypatch.delenv("VIRTUAL_ENV", raising=False)
    monkeypatch.delenv("ANSIBLE_HOME", raising=False)
    cache_dir = get_cache_dir(Path(), isolated=True)
    assert cache_dir == Path().cwd() / ".ansible"


def test_get_cache_dir_isolation_no_venv_root(monkeypatch: MonkeyPatch) -> None:
    """Test behaviors of get_cache_dir.

    Args:
        monkeypatch: Pytest fixture for monkeypatching
    """
    monkeypatch.delenv("VIRTUAL_ENV", raising=False)
    monkeypatch.delenv("ANSIBLE_HOME", raising=False)
    with (
        pytest.warns(
            UserWarning,
            match=r"Project directory /.ansible cannot be used for caching as it is not writable.",
        ),
        pytest.warns(
            UserWarning,
            match=r"Using unique temporary directory .* for caching.",
        ),
    ):
        cache_dir = get_cache_dir(Path("/"), isolated=True)
    assert cache_dir.as_posix().startswith(tempfile.gettempdir())


def test_get_cache_dir_venv_ro_project_ro(monkeypatch: MonkeyPatch) -> None:
    """Test behaviors of get_cache_dir with read-only virtual environment and read only project directory.

    Args:
        monkeypatch: Pytest fixture for monkeypatching
    """
    monkeypatch.setenv("VIRTUAL_ENV", "/")
    monkeypatch.delenv("ANSIBLE_HOME", raising=False)
    with (
        pytest.warns(
            UserWarning,
            match=r"Using unique temporary directory .* for caching.",
        ),
        pytest.warns(
            UserWarning,
            match=r"Found VIRTUAL_ENV=/ but we cannot use it for caching as it is not writable.",
        ),
        pytest.warns(
            UserWarning,
            match=r"Project directory .* cannot be used for caching as it is not writable.",
        ),
    ):
        cache_dir = get_cache_dir(Path("/etc"), isolated=True)
    assert cache_dir.as_posix().startswith(tempfile.gettempdir())



"""Pytest fixtures."""

import importlib.metadata
import json
import pathlib
import subprocess
import sys
from collections.abc import Callable, Generator
from pathlib import Path

import pytest

from ansible_compat.runtime import Runtime


@pytest.fixture
# pylint: disable=unused-argument
def runtime(scope: str = "session") -> Generator[Runtime, None, None]:  # noqa: ARG001
    """Isolated runtime fixture."""
    instance = Runtime(isolated=True)
    yield instance
    instance.clean()


@pytest.fixture
# pylint: disable=unused-argument
def runtime_tmp(
    tmp_path: pathlib.Path,
    scope: str = "session",  # noqa: ARG001
) -> Generator[Runtime, None, None]:
    """Isolated runtime fixture using a temp directory."""
    instance = Runtime(project_dir=tmp_path, isolated=True)
    yield instance
    instance.clean()


def query_pkg_version(pkg: str) -> str:
    """Get the version of a current installed package.

    :param pkg: Package name
    :return: Package version
    """
    return importlib.metadata.version(pkg)


@pytest.fixture
def pkg_version() -> Callable[[str], str]:
    """Get the version of a current installed package.

    :return: Callable function to get package version
    """
    return query_pkg_version


class VirtualEnvironment:
    """Virtualenv wrapper."""

    def __init__(self, path: Path) -> None:
        """Initialize.

        :param path: Path to virtualenv
        """
        self.project = path
        self.venv_path = self.project / "venv"
        self.venv_bin_path = self.venv_path / "bin"
        self.venv_python_path = self.venv_bin_path / "python"

    def create(self) -> None:
        """Create virtualenv."""
        cmd = [str(sys.executable), "-m", "venv", str(self.venv_path)]
        subprocess.check_call(args=cmd)
        # Install this package into the virtual environment
        self.install(f"{__file__}/../..")

    def install(self, *packages: str) -> None:
        """Install packages in virtualenv.

        :param packages: Packages to install
        """
        cmd = [str(self.venv_python_path), "-m", "pip", "install", *packages]
        subprocess.check_call(args=cmd)

    def python_script_run(self, script: str) -> subprocess.CompletedProcess[str]:
        """Run command in project dir using venv.

        :param args: Command to run
        """
        proc = subprocess.run(
            args=[self.venv_python_path, "-c", script],
            capture_output=True,
            cwd=self.project,
            check=False,
            text=True,
        )
        return proc

    def site_package_dirs(self) -> list[Path]:
        """Get site packages.

        :return: List of site packages dirs
        """
        script = "import json, site; print(json.dumps(site.getsitepackages()))"
        proc = subprocess.run(
            args=[self.venv_python_path, "-c", script],
            capture_output=True,
            check=False,
            text=True,
        )
        dirs = json.loads(proc.stdout)
        if not isinstance(dirs, list):
            msg = "Expected list of site packages"
            raise TypeError(msg)
        sanitized = list({Path(d).resolve() for d in dirs})
        return sanitized


@pytest.fixture(scope="module")
def venv_module(tmp_path_factory: pytest.TempPathFactory) -> VirtualEnvironment:
    """Create a virtualenv in a temporary directory.

    :param tmp_path: pytest fixture for temp path
    :return: VirtualEnvironment instance
    """
    test_project = tmp_path_factory.mktemp(basename="test_project-", numbered=True)
    venv_ = VirtualEnvironment(test_project)
    venv_.create()
    return venv_
